import ArrayKeyedMap from 'array-keyed-map'

import {
   Implementations, isPlainFunction, isPlainObject, onType
} from './helpers.js'
import {bootstrapTypes, Returns, whichType, Type} from './Type.js'
import {matched, needsCollection, Passthru} from './TypePatterns.js'

export class TypeDispatcher {
   constructor(...specs) {
      this._types = {} // stores all the types registered in this dispatcher
      this._implementations = {} // maps key to list of [pattern, result] pairs
      this._dependencies = {} // maps key to a map from type vectors to...
      // ...a set of [key, types] that depend on it.
      this._behaviors = {} // maps key to a map from type vectors to results
      this._fallbacks = {} // maps key to a catchall result
      this.merge({types: () => this._types})
      this.merge(bootstrapTypes) // bootstrap the instance
      for (const spec of specs) this.merge(spec)
   }

   // Only modify a TypeDispatcher via merge! Otherwise dependencies will
   // not necessarily be updated properly. Actually I may have it set up so
   // that's the only way to modify a TypeDispatcher... not sure, should
   // probably test it.
   merge(spec) {
      if (!spec) return
      if (typeof spec != 'object') {
         throw new TypeError(
            `TypeDispatcher specifications must be objects, not '${spec}'.`)
      }
      for (const key in spec) {
         let val = spec[key]
         if (val instanceof Type) {
            // The design here is that we have set up the `types` property to
            // be watched like any other, so the following assignment will
            // cause any behaviors that depend on the type named `key` in the
            // list of types to be cleared out automagically. Seems to work
            // so far.
            this.types[key] = val
            val.name = key
            continue
         }
         if (typeof val === 'function') {
            val = onType(Passthru, val)
         }
         if (val instanceof Implementations) {
            if (!(key in this)) {
               // Need to "bootstrap" the item:
               // We initially define it with a temporary getter, only
               // because we don't know whether ultimately it will produce
               // a function or a non-callable value, and the "permanent"
               // getter we want depends on which it turns out to be. This
               // situation means it's not supported to replace a key
               // corresponding to a method, after it's been fetched once,
               // with a definition that produces a non-callable value;
               // the TypeDispatcher will go on returning a function, and
               // even if you call that function, it will throw an error
               // when it tries to call the non-callable value.
               // Conversely, if you try to replace a key corresponding
               // to a non-callable value, after it's been fetched once,
               // with a function, it will work, but it will not be able to
               // perform type dispatch on that function.
               Object.defineProperty(this, key, {
                  enumerable: true,
                  configurable: true,
                  get: () => {
                     let tryValue
                     try {
                        tryValue = this.resolve(key, [])
                     } catch {
                        // Has no value for the empty type list, so therefore
                        // it must be a method, as there is no way to supply
                        // any types for a non-function value. Hence, we can
                        // just make tryValue any plain function, since it is
                        // never actually used, just its type analyzed.
                        tryValue = () => undefined
                     }
                     // Redefine the property according to what sort of
                     // entity it is:
                     if (isPlainFunction(tryValue)) {
                        const thisTypeOf = whichType(this.types)
                        // the usual case: a method of the dispatcher
                        const standard = (...args) => {
                           const types = args.map(thisTypeOf)
                           return this.resolve(key, types)(...args)
                        }
                        standard.resolve = (types) => this.resolve(key, types)
                        standard.isDispatcher = true
                        Object.defineProperty(this, key, {
                           enumerable: true,
                           configurable: true,
                           get: () => standard
                        })
                        return standard
                     }

                     if (typeof tryValue === 'object') {
                        if (!('resolve' in tryValue)) {
                           tryValue.resolve = types => this.resolve(key, types)
                        }
                        const get = () => {
                           if (this._dependencies[key]?.get([])?.size) {
                              return DependencyWatcher(tryValue, [key], this)
                           }
                           return tryValue
                        }
                        Object.defineProperty(this, key, {
                           enumerable: true,
                           configurable: true,
                           get
                        })
                        return get()
                     }

                     Object.defineProperty(this, key, {
                        enumerable: true,
                        configurable: true,
                        get: () => tryValue
                     })
                     return tryValue
                  }
               })

               // Finally, initialize the other data for this key:
               this._implementations[key] = []
               this._behaviors[key] = new ArrayKeyedMap()
               this._dependencies[key] = new ArrayKeyedMap()
            }

            // Now add all of the patterns of this implementation:
            for (const [pattern, result] of val.patterns) {
               if (pattern === Passthru) {
                  if (key in this._fallbacks) this._disengageFallback(key)
                  this._fallbacks[key] = result
               } else {
                  this._clearBehaviorsMatching(key, pattern)
                  // if it happens the same pattern is already in the
                  // implementations, remove it
                  const imps = this._implementations[key]
                  const have = imps.findIndex(elt => pattern.equal(elt[0]))
                  if (have >= 0) imps.splice(have, 1)
                  this._implementations[key].unshift([pattern, result])
               }
            }
            continue
         }

         if (isPlainObject(val)) {
            this.merge(val)
            continue
         }

         // install value as a catchall value
         this.merge({[key]: onType(Passthru, val)})
      }
   }

   _addToDeps(key, subkey) {
      const depMap = this._dependencies[key]
      if (!depMap.has(subkey)) depMap.set(subkey, new Set())
      const depSet = depMap.get(subkey)
      for (const pair of this.resolve._genDepsOf) depSet.add(pair)
   }

   // produces and returns a function that takes a list of arguments and
   // transforms them per the given template, to massage them into the form
   // expected by a behavior associated with the TypePattern that produced
   // the template.
   _generateCollectFunction(template, state = {pos: 0}) {
      const extractors = []
      for (const elt of template) {
         if (Array.isArray(elt)) {
            extractors.push(this._generateCollectFunction(elt, state))
         } else {
            const from = state.pos++
            if ('actual' in elt) { // incorporate conversion
               let convert = elt.convertor
               if (!convert.returns) { // it's a factory that produces convert
                  convert = convert(this, elt.actual)
               }
               extractors.push(args => convert(args[from]))
            } else extractors.push(args => args[from])
         }
      }
      return args => extractors.map(f => f(args))
   }

   resolve(key, types) {
      if (!(key in this)) {
         throw new ReferenceError(`no method or value for key '${key}'`)
      }

      if (this.resolve._genDepsOf?.length) this._addToDeps(key, types)

      const behave = this._behaviors[key]
      // Return the cached resolution if it's there
      if (behave.has(types)) return behave.get(types)

      // Otherwise, perform the resolution and cache the result
      const imps = this._implementations[key]
      let needItem = true
      let item
      let template
      if (imps.length) {
         for (const options of [{}, {convert: true}]) {
            let pattern
            for ([pattern, item] of imps) {
               let finalIndex
               ;[finalIndex, template] = pattern.match(types, options)
               if (finalIndex === types.length) {
                  needItem = false
                  break
               }
            }
            if (!needItem) break
         }
      }
      if (needItem && key in this._fallbacks) {
         needItem = false
         item = this._fallbacks[key]
         template = types
      }
      if (needItem) {
         throw new TypeError(`no matching definition of '${key}' on '${types}'`)
      }
      // If this key is producing a non-function value, we're done
      if (!isPlainFunction(item)) {
         behave.set(types, item)
         return item
      }

      // item is a function, either a direct behavior or
      // a factory. We have to use it to build the final behavior
      // First set up to record dependencies
      if (!('_genDepsOf' in this.resolve)) {
         this.resolve._genDepsOf = []
      }

      let theBehavior = () => undefined
      this.resolve._genDepsOf.push([key, types]) // Important: make sure
      // not to return without popping _genDepsOf
      if (!('returns' in item)) {
         // looks like a factory
         try {
            theBehavior = item(
               DependencyRecorder(this, [], this), matched(template))
         } catch {
            // Oops, didn't work as a factory, so guess we were wrong.
            // Just make it the direct value for this key on these types:
            theBehavior = item
         }
      } else theBehavior = item

      let finalBehavior = theBehavior
      if (typeof theBehavior === 'function') {
         const returning = theBehavior.returns
         if (!returning) {
            throw new TypeError(
               `No return type specified for ${key} on ${types}`)
         }
         if (theBehavior.length && needsCollection(template)) {
            // have to wrap the behavior to collect the actual arguments
            // in the way corresponding to the template. Generating that
            // argument transformer may generate more dependencies.
            const morph = this._generateCollectFunction(template)
            finalBehavior =
               Returns(returning, (...args) => theBehavior(...morph(args)))
         }
      }

      this.resolve._genDepsOf.pop()  // OK, now it's safe to return
      behave.set(types, finalBehavior)
      return finalBehavior
   }

   // Method called to invalidate a set of behaviors
   // I think all it has to do is throw them out; they should be
   // regenerated properly by resolve, if all is well.
   _invalidate(behaveSet) {
      for (const [key, types] of behaveSet) this._behaviors[key].delete(types)
   }

   _disengageFallback(key) {
      // We need to find all of the behaviors that currently rely on the
      // fallback, invalidate their dependencies, and remove them.
      const fallTypes = []
      const behs = this._behaviors[key]
      const imps = this._implementations[key]
      const deps = this._dependencies[key]
      for (const types of behs.keys()) {
         let fallsback = true
         for (const [pattern] of imps) {
            const [finalIndex] = pattern.match(types)
            if (finalIndex === types.length) {
               fallsback = false
               break
            }
         }
         if (fallsback) fallTypes.push(types)
      }
      for (const types of fallTypes) {
         const depSet = deps?.get(types)
         if (depSet?.size) this._invalidate(depSet)
         behs.delete(types)
      }
   }

   _clearBehaviorsMatching(key, pattern) {
      // like disengageFallback, just we have the offending pattern:
      const behs = this._behaviors[key]
      const deps = this._dependencies[key]
      const patTypes = behs.keys().filter(types => {
         const [finalIndex] = pattern.match(types)
         return finalIndex === types.length
      })
      for (const types of patTypes) {
         const depSet = deps?.get(types)
         if (depSet?.size) this._invalidate(depSet)
         behs.delete(types)
      }
   }
}

// Proxy that traps accesses and records dependencies on them
const DependencyRecorder = (object, path, repo) => new Proxy(object, {
   get(target, prop, receiver) {
      const result = Reflect.get(target, prop, receiver)
      // pass resolve calls through, since we record dependencies therein:
      if (prop === 'resolve'
          || (typeof result === 'function' && 'isDispatcher' in result)
      ) {
         return result
      }

      // OK, it's not a method on a TypeDispatcher, it's some other kind of
      // value. So first record the dependency on prop at this path:
      const newPath = path.slice()
      newPath.push(prop)
      const key = newPath[0]
      const subpath = newPath.slice(1)
      repo._addToDeps(key, subpath)
      // Now, if the result is an object, we may need to record further
      // dependencies on its properties (e.g. math.config.predictable)
      // So proxy the return value, except for types, which must maintain
      // strict referential identity:
      if (typeof result === 'object' && !(result instanceof Type)) {
         return DependencyRecorder(result, newPath, repo)
      } else return result
   }
})
      
// The flip side: proxy that traps setting properties and invalidates things
// that depend on them:
const DependencyWatcher = (object, path, repo) => new Proxy(object, {
   set(target, prop, value, receiver) {
      // First see if this setting has any dependencies:
      const newPath = path.slice()
      newPath.push(prop)
      const key = newPath.shift()
      const depSet = repo._dependencies[key]?.get(newPath)
      if (depSet?.size) {
         // It does. So if we are changing it, invalidate them:
         const oldValue = Reflect.get(target, prop, receiver)
         if (value !== oldValue) repo._invalidate(depSet)
      }
      // Now we can just perform the setting
      return Reflect.set(target, prop, value, receiver)
   },
   get(target, prop, receiver) {
      // Only thing we need to do is push the watching down
      const result = Reflect.get(target, prop, receiver)
      if (typeof result === 'object' && !(result instanceof Type)) {
         const newPath = path.slice()
         newPath.push(prop)
         return DependencyWatcher(result, newPath, repo)
      }
      return result
   }
})
