import {Type, Undefined} from './Type.js'

export class TypePattern {
   match(typeSequence, options={}) {
      throw new Error('Specific TypePatterns must implement match')
   }
   sampleTypes() {
      throw new Error('Specific TypePatterns must implement sampleTypes')
   }
   equal(other) {return other.constructor === this.constructor}
}

class MatchTypePattern extends TypePattern {
   constructor(typeToMatch) {
      super()
      this.type = typeToMatch
   }
   match(typeSequence, options={}) {
      const position = options.position ?? 0
      const actual = typeSequence[position]
      if (position < typeSequence.length) {
         if (actual === this.type) return [position + 1, actual]
         if (options.convert) {
            for (const [pattern, convertor] of this.type.from.patterns) {
               const [pos] = pattern.match([actual])
               if (pos === 1) {
                  return [position + 1, {actual, convertor, matched: this.type}]
               }
            }
         }
      }
      return [-1, Undefined]
   }
   sampleTypes() {return [this.type]}
   equal(other) {return super.equal(other) && this.type === other.type}
}

class SequencePattern extends TypePattern {
   constructor(itemsToMatch) {
      super()
      this.patterns = itemsToMatch.map(pattern)
   }
   match(typeSequence, options={_internal: true}) {
      options = options._internal
         ? options
         : Object.assign({_internal: true}, options)
      options.position ??= 0
      const matches = []
      for (const pat of this.patterns) {
         const [newPos, newMatch] = pat.match(typeSequence, options)
         if (newPos < 0) return [-1, Undefined]
         options.position = newPos
         matches.push(newMatch)
      }
      return [options.position, matches]
   }
   sampleTypes() {
      return this.patterns.map(pat => pat.sampleTypes()).flat()
   }
   equal(other) {
      return super.equal(other)
         && this.patterns.length === other.patterns.length
         && this.patterns.every((elt, ix) => elt.equal(other.patterns[ix]))
   }
}

export const pattern = patternOrSpec => {
   if (patternOrSpec instanceof TypePattern) return patternOrSpec
   if (patternOrSpec instanceof Type) {
      return new MatchTypePattern(patternOrSpec)
   }
   if (Array.isArray(patternOrSpec)) return new SequencePattern(patternOrSpec)
   throw new TypeError(`Can't interpret '${patternOrSpec}' as a type pattern`)
}

class AnyPattern extends TypePattern {
   match(typeSequence, options={}) {
      const position = options.position ?? 0
      return position < typeSequence.length
         ? [position + 1, typeSequence[position]]
         : [-1, Undefined]
   }
   sampleTypes() {return [Undefined]}
}

export const Any = new AnyPattern()

class OptionalPattern extends TypePattern {
   constructor(item) {
      super()
      this.pattern = pattern(item)
   }
   match(typeSequence, options={_internal: true}) {
      options = options._internal
         ? options
         : Object.assign({_internal: true}, options)
      options.position ??= 0
      const matches = []
      const [newPos, newMatch] = this.pattern.match(typeSequence, options)
      if (newPos >= 0) {
         options.position = newPos
         matches.push(newMatch)
      }
      return [options.position, matches]
   }
   sampleTypes() {return []}
   equal(other) {
      return super.equal(other) && this.pattern.equal(other.pattern)
   }
}

export const Optional = item => new OptionalPattern(item)

class MultiPattern extends TypePattern {
   constructor(item) {
      super()
      this.pattern = pattern(item)
   }
   match(typeSequence, options={_internal: true}) {
      options = options._internal
         ? options
         : Object.assign({_internal: true}, options)
      options.position ??= 0
      const matches = []
      while (true) {
         const [newPos, newMatch] = this.pattern.match(typeSequence, options)
         if (newPos < 0) return [options.position, matches]
         options.position = newPos
         matches.push(newMatch)
      }
   }
   sampleTypes() {return []}
   equal(other) {
      return super.equal(other) && this.pattern.equal(other.pattern)
   }
}

export const Multiple = item => new MultiPattern(item)

// Like Multiple(Any) except leaves the argument list alone; it doesn't
// chunk it into a single Array of all arguments
class PassthruPattern extends TypePattern {
   match(typeSequence, options={}) {
      const position = options.position ?? 0
      return [typeSequence.length, typeSequence.slice(position)]
   }
   sampleTypes() {return []}
}

export const Passthru = new PassthruPattern()

// returns the template just of matched types, dropping any actual types
export const matched = (template) => {
   if (Array.isArray(template)) return template.map(matched)
   return template.matched ?? template
}

// checks if the template is just pass-through or needs collection
export const needsCollection = (template) => {
   if (Array.isArray(template)) {
      return template.some(
         elt => Array.isArray(elt) || needsCollection(elt))
   }
   return 'actual' in template
}
